// SPDX-License-Identifier: BSD-2-Clause
/* Copyright (C) 2014 - 2020 Intel Corporation. */

#pragma once
#ifdef __cplusplus
extern "C" {
#endif

#include <memkind.h>

/*
 * Header file for the gigabyte TLB memkind operations.
 *
 * All function declarations has been moved to memkind_deprecated.h
 * because of end of GBTLB support.
 *
 * API standards are described in memkind(3) man page.
 */

#ifdef __cplusplus
}
#endif
