// SPDX-License-Identifier: BSD-2-Clause
/* Copyright (C) 2018 - 2020 Intel Corporation. */
#pragma once
#include "memkind.h"

///
/// \brief Mockup structure for PMEM kind to use allocator_perf_tool engine
///
extern memkind_t MEMKIND_PMEM_MOCKUP;
